/* 
 * MulticastCapsule.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  $Id: MulticastCapsule.Java,v 1.3 2000/02/08 23:42:56 tullmann Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_MULTICAST_APP_PKG;

import JANOS_ANTSR_CORE_PKG.*;

/**
 * Multicast capsule processing.  These correspond to the multicast data
 * packets.
 *
 * @author David Wetherall
 */
public class MulticastCapsule
	extends Capsule
{
	final private static byte[] MID
		= findMID(JANOS_ANTSR_MULTICAST_APP_PKG_S
			  + ".MulticastCapsule");
	protected byte[] mid() { return MID; }
	
	final private static byte[] PID
		= findPID(JANOS_ANTSR_MULTICAST_APP_PKG_S
			  + ".MulticastCapsule");
	protected byte[] pid() { return PID; }
	
	PACKAGE int group;	// The multicast group (address).
	PACKAGE short destport;	// Destination port
	PACKAGE ByteArray data;	// The actual data for this capsule.
	
	/**
	 * Return the length of this capsule.
	 */
	public int length() {
		int s = Xdr.INT + Xdr.SHORT + Xdr.BYTEARRAY(data);
		return super.length() + s;
	}
	
	
	/**
	 * Encode this capsule.
	 */
	public Xdr encode() {
		Xdr xdr = super.encode();
		
		xdr.PUT(group);
		xdr.PUT(destport);
		xdr.PUT(data);
		
		return xdr;
	}
	
	/**
	 * Decode this capsule.
	 */
	public Xdr decode() {
		Xdr xdr = super.decode();
		
		group = xdr.INT();
		destport = xdr.SHORT();
		data = xdr.BYTEARRAY();
		
		return xdr;
	}
	
	/**
	 * Process this capsule at the node.
	 *
	 * @param n The current node.
	 */
	public boolean evaluate(Node n) {
		// Create a key and try to get our soft state.
		MulticastCapsuleCacheKey mkey
			= new MulticastCapsuleCacheKey(group, getDst());
		MulticastCapsuleCacheValue mval
			= (MulticastCapsuleCacheValue)n.getCache().get(mkey);
		n.log(0, "MulticastCapsule: Trying to get state for group "
		      + NodeAddress.toString(group)
		      + ", dest " + NodeAddress.toString(getDst())
		      + ", key " + mkey.hashCode());
		
		if (mval != null) {
			/* We found an entry. */
			if (mval.addrs != null) {
				/* Send the capsule to all nodes listed. */
				for (int i = 0; i < mval.addrs.length; i++)
					n.routeForNode(this, mval.addrs[i]);
			} else {
				/* Deliver to application. */
				n.deliverToApp(this, destport);
			}
		} else {
			/* There is no entry, we can only drop it. */
			int count = ((int)data.getByte(4)) << 24
				    | ((int)data.getByte(5)) << 16
				    | ((int)data.getByte(6)) << 8
				    | (data.getByte(7) & 0xff);
			n.log(0, "MulticastCapsule: Dropping capsule #" + count
			      + " -- Nowhere to send!");
		}
		return true;
	}
	
	/* Constructors. */
	public MulticastCapsule() { }
	public MulticastCapsule(int group, int target, short destport,
				ByteArray data) {
		this.group = group;
		this.destport = destport;
		this.data = data;
		setDst(target);
	}
}
