/* 
 *  $Id: Application.Java,v 1.2 2000/01/12 21:20:33 stack Exp $
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.CommQueue;
import JANOS_NODEOS_PKG.CommSpaceElement;

import JANOS_ANTSR_UTIL_PKG.KeyArgs;
import JANOS_ANTSR_UTIL_PKG.InvalidKeyArgsException;

IMPORT_DEBUG

/**
 * Abstract node application.  An Application is an end-point for capsules.
 * Capsules are sent and received via the Node associated with the Application.
 * <p>
 * Applications understand the '-port' KeyArg for setting their port address
 * on the Node.  If no '-port' is provided an available port will be found
 * and allocated to the Application.  getPort() will retrieve the port
 * number.  
 * <p>
 * Applications are started after the Node has completed initialization.
 *
 * @author David Wetherall
 */
public abstract class Application
	extends Entity
{
	/**
	 * A queue to accept packets for the application.  The protocol for use
	 * is to send the MID, followed by the packet bytes.
	 */
	private final CommQueue packetQueue = new CommQueue();
	
	/**
	 * Default KeyArgs for an application.  Sets port to 0.
	 */
	final public static String[] defaults = { "-port", "0" };

	/**
	 * The Node associated with this application.  All Capsules
	 * are sent via and recieved from this Node.
	 *
	 * @see #setNode()
	 * @see #getNode()
	 */
	private Node node;

	/**
	 * The default default resource limit for packets
	 * sent by an Application.
	 *
	 * @deprecated Please use DEFAULT_TTL instead.
	 */
	private final static int DEFAULTRLIMIT = 64;

	/**
	 * The default TTL for packets sent by this application.
	 */
	private final static int DEFAULT_TTL = 64;
	
	
	/**
	 * The default resource limit for Capsules sent by
	 * this application.
	 *
	 * @see #setDefaultResources(int)
	 * @see #getDefaultResources()
	 */
	private int ttl;

	public int getDefaultTTL()
	{
		return ttl;
	}
	
	public void setDefaultTTL(int ttl)
	{
		assert(ttl > 0);
		this.ttl = ttl;
	}

	/**
	 * @deprecated use getDefaultTTL()
	 */
	public int getDefaultResources()
	{
		return getDefaultTTL();
	}

	/**
	 * @deprecated use setDefaultTTL(int)
	 */
	public void setDefaultResources(int r)
	{
		setDefaultTTL(r);
	}

	public short getPort()
	{
		return port;
	}

	/**
	 * Set this application's port.  Only
	 * invoked by Node.attachApplication()
	 * if this app didn't request a specific
	 * port number.
	 */
	PACKAGE void setPort(int p)
	{
		assert(port == 0);
		assert(p < (64 * 1024));
		port = (short)p;
	}

	/**
	 * Return the Node interface customized for this application.
	 */
	public Node thisNode() {
		return node; 
	}

	/**
	 * Assign the node interface to this application.  Done
	 * only in AppBooster.run().
	 *
	 * @see AppBooster#run()
	 */
	PACKAGE void setNode(Node node)
	{
		this.node = node;
	}
  

	/**
	 * The port for this application.  This is the identifier by
	 * which remote nodes can send capsules to this application:
	 * the address determines the Node, and the port determines
	 * the application.
	 * <p>
	 * Set by the '-port' KeyArg.  Defaults to 0.
	 * <p>
	 * Is read-only after being set by '-port' (the read-only
	 * semantics are required by Node.detachApplication()).
	 * 
	 * @see Node#attachApplication()
	 * @see Node#detachApplication()
	 */
	private short port = 0;

	/**
	 * Send the given capsule with the given resource limit.
	 * <p>
	 * The capsule is primed (that is, the 'from address' and ttl
	 * are set), and then sent via the associated Node.
	 *
	 * @param cap the Capsule to send.  XXX state of Xdr?
	 * @param ttl the TTL to give the Capsule.
	 * @see Node#send(Capsule)
	 * @see #prime(int,int)
	 */
	public boolean send(Capsule cap, int ttl) {
		cap.prime(node.getAddress(), ttl); // privileged prime()
		return node.send(cap);
	}
	
	/**
	 * Send the given capsule.
	 * <p>
	 * The capsule is primed with the default TTL (this.ttl)
	 * then, the capsule is sent via the associated Node.
	 *
	 * @see #ttl
	 * @see #prime(int,int)
	 * @see Node#send(Capsule)
	 */
	public boolean send(Capsule cap) {
		CommQueue queue, replyQueue;
		boolean done = false;
		int sent = 0;

		/*
		 * XXX This kinda sucks, we need to return something so we
		 * have to setup a queue for getting the reply.
		 */
		replyQueue = new CommQueue();
		cap.prime(node.getAddress(), this.ttl); // privileged prime()
		queue = PrimordialNode.
			getProtocolQueue(TypeID.wrap(cap.pid()));
		synchronized(queue)
		{
			queue.enqueue(new CommSpaceElement(ProtocolBooster.
							   APP_CAPSULE));
			cap.enqueueCapsule(queue);
			queue.enqueue(new CommSpaceElement(replyQueue));
		}
		while(!done)
		{
			try
			{
				sent = replyQueue.waitToTake().getInt();
				done = true;
			}
			catch(InterruptedException e)
			{
			}
		}
		return (sent != 0);
	}
  
	/**
	 * Callback for packets that are delivered to this application.
	 * This does nothing, so the default behavior for the application
	 * is to drop delivered packets on the floor.
	 *
	 * @param cap the Capsule that got delivered.  XXX state of Xdr?
	 */
	public void receive(Capsule cap) {
		// Log("Application.receive() got " + cap);
	}
  
	/**
	 * Start this Application rolling.  This is invoked once by the
	 * Node after it has been initialized.
	 * <p>
	 * Note that implementations of start should just set things up.  If
	 * the application wants to do things like send packets, this
	 * function should start a new thread to do that.  Creating a GUI
	 * and associating callbacks is a good thing to do here.
	 * <p>
	 * XXX mostly stuff that could just be done in the no-arg 
	 * constructor, no?
	 */
	public abstract void start()
		throws Exception;
	
	/**
	 * Parse command line arguments for this application.
	 * Understands the -port option.
	 * <p>
	 * It is guaranteed that this method will be invoked
	 * before the application starts running (i.e. shortly
	 * after the constructor is invoked).
	 *
	 * @see #port
	 * @see ManagedObject#setArgs(utils.KeyArgs)
	 */
	public void setArgs(KeyArgs k) 
		throws InvalidKeyArgsException
	{
		k.merge(defaults);

		for (int i = 0; i < k.length(); i++) {
			if (k.key(i).equals("-port")) {
				port = (short)Integer.parseInt(k.arg(i));
				k.strike(i);
			}
		}
    
		k.noMore();
	}

	/**
	 * Return the packet queue for this application
	 */
	PACKAGE final CommQueue getPacketQueue()
	{
		return this.packetQueue;
	}
	
	/**
	 * All applications must have a no-arg constructor as the
	 * Node invokes Class.newInstance() to create the application.
	 */
	public Application()
	{
		setDefaultTTL(DEFAULT_TTL);
	}

	public String toString()
	{
		return "Application[node="
			+ this.node
			+ "; ttl="
			+ this.ttl
			+ "; port="
			+ this.port
			+ "]";
	}
}
