/* 
 *  $Id: Capsule.Java,v 1.7 2000/05/02 23:57:19 tullmann Exp $
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.CommQueue;
import JANOS_NODEOS_PKG.CommSpaceElement;
import JANOS_IO_PKG.Log;

IMPORT_DEBUG

/**
 * A capsule is a combination of a packet and its forwarding routine;
 * the forwarding routine is executed at every active node the
 * packet visits in the network.
 * <p>
 * This is an abstract, an actual Capsule is created by subclassing
 * this class and filling in the missing routines.
 * <p>
 * Subclasses must implement the mid(), pid(), and evalutate() methods.
 * <p>
 * All ANTSR capsules have the following format:
 * <pre>
 *   16-byte MethodID  (The MD5 hash over the class bytecode)
 *   4-byte logical source address
 *   4-byte logical destination address
 *   4-byte logical previous-hop address
 *   4-byte TTL field.
 *   ... Capsule-specific data ...
 * </pre>
 *
 * @author David Wetherall
 */
public abstract class Capsule {
	PACKAGE static final byte ANEPVERSION = 1;
	PACKAGE static final short ANTSTYPE = 18; // MIT ANTS type
	PACKAGE static final short ANTSRTYPE = 121;
	PACKAGE static final short SRCOPTTYPE = 1;
	PACKAGE static final short DESTOPTTYPE = 2;
	PACKAGE static final int IPV4SCHEME = 1;
	PACKAGE static final int ANTSRVERSION = 0x00010004;
	PACKAGE static final int HEADERLENGTH =
		Xdr.TYPEID + Xdr.INT + Xdr.INT + Xdr.INT + Xdr.INT;

	public static final int MAX_TTL = 255;

	/**
	 * Return the MethodID associated with this capsule.
	 * <p>
	 * See <a href="On_IDs.txt">On_IDs.txt</a> for more info on
	 * MethodIDs.
	 *
	 * @see InChannel#findMID(java.lang.String)
	 * @return the byte[] representing the MD5 checksum.
	 */
	protected abstract byte[] mid();

	/**
	 * Return the ProtocolID associated with this capsule.
	 * <p>
	 * See <a href="On_IDs.txt">On_IDs.txt</a> for more info on
	 * ProtocolIDs.
	 *
	 * @see CapsuleClass#findPID(java.lang.String)
	 * @return the byte[] representing the MD5 checksum.
	 */
	protected abstract byte[] pid();

	/**
	 * All capsules originate from some source node. 
	 * This is the 32-bit IP address of that source.
	 * <p>
	 * The source is set implicitly by the node that
	 * injects the capsule into the network and is
	 * never changed after that.
	 * <p>
	 * XXX what is this used for by the Node?  Or is
	 * it just a convenience for subclasses?  If that's
	 * all then nuke it.
	 *
	 * @see #getSrc()
	 */
	private int src;

	/**
	 * The destination for this capsule.  Forwarding code
	 * may read/write this field.
	 *
	 * @see #getDst()
	 * @see #setDst()
	 */
	private int dst;

	/**
	 * 32-bit IP address of previous active node.  This
	 * is used by the code loading protocol to dynamically find
	 * the code to process a new packet type.
	 * <p>
	 * XXX security implications of giving this package access?
	 *
	 * @see #getPrevious()
	 * @see #setPrevious()
	 */
	private int previous; // accessed directly by channel


	/**
	 * The xdr stream representing the buffer that
	 * underlies a Capsule.
	 * <p>
	 * For incoming packets,
	 * this field is set by the Protocol receive()
	 * function.
	 * <p>
	 * For outgoing packets, this field is set
	 * by the encode() method which creates a new
	 * Xdr.
	 */
	private Xdr xdr;

	/**
	 * Time to live for a packet.  Used to control
	 * packet propagation.
	 * <p>
	 * XXX a byte would suffice?
	 *
	 * @see getTTL()
	 */
	private int ttl;

	/**
	 * Create a dull capsule.  
	 */
	public Capsule() 
	{ 
	}

	/**
	 * Get the 32-bit IP address of the active node
	 * this packet just came from.  Used to find
	 * code to process this packet.
	 */
	public int getPrevious()
	{
		return previous;
	}

	/**
	 * Set the 32-bit IP address of the active node
	 * this packet "came from".  Used by the outgoing
	 * channel encode() operation.
	 *
	 * @see Channel#encode(Capsule)
	 */
	PACKAGE void setPrevious(int previous) 
	{
		this.previous = previous;
	}
	

	/**
	 * Return the TTL-like counter for this capsule.
	 *
	 * @return a TTL value for this Capsule.
	 */
	public int getTTL()
	{
		return ttl;
	}

	/**
	 * Access point to initialze capsule methodIDs.  Given
	 * Capsule class name, returns the appropriate MID.
	 * <p>
	 * XXX its probably bad that the actual underlying
	 * byte[] is returned and not a copy...  Who uses this?
	 *
	 * @param name the name of the Capsule class
	 * @return the MD5 checksum (MID) associated with the given Capsule class.
	 */
	public static byte[] findMID(String name)
	{
		// redirect to the current channel
		// return InChannel.currentInChannel().findMID(name);

		// XXX return a *copy* of the byte array???

		assert(Node.currentNode());

		TypeID t = Node.currentNode().findMID(name);
		if (t == null)
			return null;
		return t.unwrap();
	}

	/**
	 * Access point to initialze capsule protocolID.  Given
	 * Capsule class name, returns the appropriate PID.
	 *
	 * @param name the name of the Capsule class
	 * @return the MD5 checksum (PID) associated with the given Capsule class.
	 * @see InChannel#findPID(java.lang.String)
	 */
	public static byte[] findPID(String name)
	{
		// redirect to the current channel
		// return InChannel.currentInChannel().findPID(name);

		// XXX return a *copy* of the byte array???

		assert(Node.currentNode());
		
		TypeID t = Node.currentNode().findPID(name);
		if (t == null)
			return null;
		return t.unwrap();
	}

	/**
	 * Find a given extension.  Given the name for an extension,
	 * return the local "index" of that extension.
	 * <p>
	 * XXX This interface will need to be replaced to allow
	 * extensions to be referenced counted (and unloaded).
	 *
	 * @param name The name of the extension to load.
	 * @return The local "index" of the given extension.
	 * @see Extension#findNamedExtension(java.lang.String)
	 * @see Node#getExtension(int)
	 */
	public static int findExtension(String name)
	{
		return Extension.findNamedExtension(name);
	}

	/**
	 * The forwarding routine for the Capsule.
	 * This is invoked by the router when a capsule is
	 * received.
	 * <p>
	 * This method should not access other Java classes,
	 * and should only interact with the local system
	 * through the local node interface.  This routine
	 * should "terminate quickly."
	 * <p>
	 * XXX perhaps default to forwarding the Capsule?
	 *
	 * @param node the local node interface
	 * @return an indicationof whether the forwarding was
	 * successful.
	 */
	abstract public boolean evaluate(Node node);
  
	/**
	 * Return the source IP address of this capsule.
	 *
	 * @return the source IP address of this capsule.
	 */
	public int getSrc()
	{
		return src;
	}

	/**
	 * Return the destination IP address of this capsule.
	 *
	 * @return the destination IP address of this capsule.
	 */
	public int getDst()
	{
		return dst;
	}

	/**
	 * Set the destination IP address of this capsule.
	 *
	 * @param dst the destination IP address of this capsule.
	 */
	public void setDst(int dst)
	{
		this.dst = dst;
	}

	/**
	 * Return the length of the capsule header that
	 * all capsule encodings will include.
	 *
	 * @return the length of the capsule header
	 */
	public int length()
	{
		return HEADERLENGTH;
	}
  
	/**
	 * Encode this capsule's header into a new Xdr.  Return the
	 * new Xdr.
	 * <p>
	 * Because this header must go first in the Xdr, this
	 * encoder unilaterally sets the xdr buffer pointer
	 * to the beginning of the buffer and writes its
	 * data there.
	 * <p>
	 * The format of the Xdr is:
	 * <pre>
	 * MethodID         (16 bytes)
	 * src address      ( 4 bytes)
	 * dst address      ( 4 bytes)
	 * previous address ( 4 bytes)
	 * TTL              ( 4 bytes)
	 * <pre>
	 *
	 * @return the new Xdr representing this capsule.
	 * @see #xdr 
	 */
	public Xdr encode()
	{
		xdr.PUT(this.mid(), 0, Xdr.TYPEID);
		xdr.PUT(src);
		xdr.PUT(dst);
		xdr.PUT(previous);
		xdr.PUT(ttl);
		return xdr;
	}

	/**
	 * Encode an outgoing capsule.
	 * <p>
	 * This writes the ANEP and ANTSR header for an outgoing capsule in its
	 * associated Xdr.  A new Xdr is created if the Capsule doesn't already
	 * have one.  Then, the per-capsule encoding is called.
	 * <p>
	 * This function should be invoked immediately before pushing
	 * the capsule out the door.
	 * <p>
	 * XXX: rename to 'encodeForSend()'
	 * 
	 * @param node the node interface 
	 * @return Xdr returns Xdr in which the remainder of the capsule should be encoded.
	 * @exception SecurityException if the source is not defined in the capsule
	 * @exception ResourceLimitException if the capsule has insufficent resources
	 */
	public Xdr encodeCapsuleForSend(Node node)
	{
		short packetLen;

		// source, etc. must be set already
		if (this.getSrc() == 0)
		{
			panic("SecurityException: source not set in encode for " + this);// XXX throw a real exception.
		}
        
		// previous is updated for use by the code loading
		this.setPrevious(node.getAddress()); 
 
		// ttl decremented here just before we leave
		if (this.ttl-- <= 0)
		{
			throw new ResourceLimitException(); // XXX
		}

		Xdr xdr;
    
		packetLen = (short)(32 + Xdr.INT + this.length());

		// This could blow up if this.length() is a lie.
		xdr = new Xdr(packetLen);

		// Insert Header
		xdr.PUT(ANEPVERSION);
		xdr.PUT((byte)0);  // Forward packet if TypeID unknown
		xdr.PUT(ANTSRTYPE);

		// Header with src and dest options = 8 32-bit words
		xdr.PUT((short)8); // header len
		xdr.PUT(packetLen);
		
		xdr.PUT(DESTOPTTYPE); // destination option
		xdr.PUT((short)3);
		xdr.PUT(IPV4SCHEME);
		xdr.PUT(this.getDst());
		
		xdr.PUT(SRCOPTTYPE); // source option
		xdr.PUT((short)3);
		xdr.PUT(IPV4SCHEME);
		xdr.PUT(this.getSrc());
		
		xdr.PUT(ANTSRVERSION);

		assert(xdr.index() == ((1 + 1 + 2) + (2 + 2) + (2 + 2 + 4 + 4) + (2 + 2 + 4 + 4) + (4)));
		
		this.xdr = xdr;
		this.encode();

		assert(xdr.index() == xdr.length()); // XXX indicates packet changed during encode

		return xdr;
	}

	PACKAGE void decodeWith(Xdr xdr)
	{
		assert(this.xdr == null);
		this.xdr = xdr;
		this.decode();
	}

	/**
	 * Decode a capsule header from this Capsule's
	 * xdr field.
	 *
	 * @return the new Xdr based on the current xdr.
	 */
	public Xdr decode()
	{
		// Caller skip'd over the MID field.  If there was one...

		// pull out rest of ANTSR headers.
		// XXX redundant with ANEP options...
		src = xdr.INT();
		dst = xdr.INT();
		previous = xdr.INT();
		ttl = xdr.INT();

		// return Xdr so that subclasses can continue decode.
		return xdr;
	}

	/**
	 * Peek at the 'previous' field of buf.
	 * For bootstrapping capsule code loading.
	 *
	 * @param buf the Buffer to peek into.
	 * @return in the value of the 'previous' field in buf.
	 */
	PACKAGE final static int peekPrevious(BufferHandle buf)
	{
		byte[] tmpBuf = new byte[Xdr.INT];

		// XXX HACK GAK COUGH, when BufferHandle is brought up to speed, fix this...
    
		// Copy 1 INT after skipping a TYPEID and two INTs.
		buf.getBytes(Xdr.TYPEID + Xdr.INT + Xdr.INT,
			     tmpBuf, 0, Xdr.INT);
		Xdr xdr = new Xdr(tmpBuf, 0);
		return xdr.INT();
	}

	/**
	 * Prime this capsule for transmission. Use the given parent
	 * Capsule to prime the src and ttl of this capsule.
	 * <p>
	 * Does not maintain zero-sum ttl.  This capsule
	 * gets parent's ttl minus 1.  Parent loses one hop.
	 * 
	 * @param parent the capsule to take src and ttl (minus 1)
	 * from.
	 */
	public final void prime(Capsule parent)
	{
		int newTtl = --parent.ttl;

		if (newTtl <= 0)
			throw new ResourceLimitException(); // XXX

		prime(parent.src, newTtl);
	}

	/**
	 * Priveleged prime operation.  Fill in this capsule's
	 * src and ttl fields directly.  The given ttl is
	 * capped at MAX_TTL
	 * <p>
	 * This is package-access only, otherwise it would be
	 * a security hole.
	 *
	 * @param src the source address for the new packet.
	 * @param ttl the tll for the new packet.
	 * @see #MAX_TTL
	 */
	PACKAGE final void prime(int src, int ttl)
	{
		this.src = src;
		if (ttl > MAX_TTL)
			ttl = MAX_TTL;
		this.ttl = ttl;
	}

	/**
	 * Package up this capsule and dump it onto the queue for another
	 * application
	 */
	PACKAGE final void enqueueCapsule(CommQueue queue)
	{
		if( this.xdr == null )
		{
			short packetLen;

			packetLen = (short)(32 + Xdr.INT + this.length());
			this.xdr = new Xdr(packetLen);
		}
		else
			this.xdr.reset();
		
		byte packetClass[] = this.mid();
		byte packet[] = this.encode().getBytes();

		synchronized(queue)
		{
			queue.enqueue(new CommSpaceElement(packetClass));
			queue.enqueue(new CommSpaceElement(packet));
		}
	}
	
	/**
	 * Produce a capsule from data pushed through a CommQueue by
	 * enqueueCapsule
	 */
	PACKAGE final static Capsule dequeueCapsule(Node node, CommQueue queue)
	{
		byte[] midbytes = null;
		byte[] packet = null;
		Capsule retval = null;
		boolean done = false;
		
		while( !done )
		{
			try
			{
				/*
				 * Get the MID and actual packet from the
				 * queue, we're being a bit anal here since its
				 * possible we could get interrupted during the
				 * waitToTake and bad things would happen.
				 */
				if( midbytes == null )
					midbytes = queue.waitToTake().
						getByteArray();
				if( packet == null )
					packet = queue.waitToTake().
						getByteArray();
				
				/* Lookup the class for this MID */
				TypeID mid = TypeID.wrap(midbytes);
				CapsuleClass cc = node.lookupCapsuleClass(mid);
				
				if( cc == null )
				{
					node.log(Log.ERR,
						 "Couldn't resolve MID ("
						 + mid
						 + ") to a capsule class");
					done = true;
				}
				/*
				 * Convert the packet bytes into an XDR
				 * for decoding
				 */
				Xdr xdr = new Xdr(packet);
				
				try
				{
					Class cl = cc.cl;
					
					done = true;
					/* Skip header junk */
					xdr.skip(16);
					retval = Capsule.create(cl, xdr);
				}
				catch(Throwable th)
				{
					node.log(Log.ERR,
						 "There was an error in "
						 + "decoding a packet");
					node.logStackTrace(Log.ERR, th);
				}
			}
			catch(InterruptedException e)
			{
				/* Oops... */
			}
		}
		return retval;
	}
	
	/**
	 * String representation of a Capsule.  Includes the
	 * class name, source and destination.
	 */
	public String toString()
	{
		return this.getClass().getName() +
			"[" +
			"s=" +NodeAddress.toString(src)+ ";" +
			"d=" +NodeAddress.toString(dst)+ 
			"]";
	}

	/**
	 * Create a new Capsule from the given Xdr.  The
	 * xdr will be decoded into an instance of the
	 * given class which must be a subclass of Capsule.
	 *
	 * @param cl the subtype of capsule to create
	 * @param xdr the encoding of the actual capsule
	 * @return the new Capsule instance.
	 */
	public static Capsule create(Class cl, Xdr xdr)
		throws Exception
	{
		Capsule c;
		
		c = (Capsule)(cl.newInstance());
		c.xdr = xdr;
		c.decode();
		return c;
	}
}
