/* 
 * MIDIDRequestCapsule.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_IO_PKG.Log;

import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * MID ID Request capsule.
 * <p>
 * The MIDIDRequestCapsule adds one field to the base Capsule:
 * <ul>
 *  <li> the typeid of the unknown MID (this.unknownMID)
 * </ul>
 *
 * <p> Pat said to make a note that this capsule could be a flow
 * control message for data packets.  It says, "woah, stop sending
 * data until the code is loaded".  But we don't take advantage of
 * this yet.
 *
 * @see DLProtocol
 * @see MIDIDResponseCapsule
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE final class MIDIDRequestCapsule
  extends Capsule 
{
	// WARNING: this is a builtin capsule, whose version number
	// must be manually kept in synch with  its functionality
	public final static String VERSION = "JANOS MIDIDRequestCapsule July 27, 1999";

	PACKAGE final static byte[] MID = findMID(JANOS_ANTSR_CORE_PKG_S + ".MIDIDRequestCapsule");
	private final static byte[] PID = findPID(JANOS_ANTSR_CORE_PKG_S + ".MIDIDRequestCapsule");

	protected byte[] mid()
	{
		return MID;
	} 

	protected byte[] pid()
	{
		return PID;
	}

	/**
	 * The TypeID of the unknown MID this Capsule represents.
	 * This is set only by the constructor (for outgoing
	 * MIDIDRequestCapsules) or by the decode() function (for
	 * incoming MIDIDRequestCapsules).
	 */
	private TypeID unknownMID;

	public int length()
	{
		return super.length() + Xdr.TYPEID;
	}
  
	public Xdr encode() 
	{
		Xdr xdr = super.encode();
		xdr.PUT(unknownMID);

		return xdr;
	}

	public Xdr decode() 
	{
		Xdr xdr = super.decode();
		unknownMID = xdr.TYPEID();

		return xdr;
	}

	/**
	 * Evaluate a MIDIDRequestCapsule.  Based on the unknownMID stored
	 * in this capsule, lookup the CapsuleClass on the local node.
	 * If the CapsuleClass is not available drop the request on
	 * the floor.  Otherwise, send a MIDIDResponseCapsule with the
	 * Protocol identity of the MID.
	 * <p>
	 * If the MIDIDRequestCapsule is for a builtin type, just drop
	 * the request.
	 *
	 * @see CodeGroup
	 */
	public boolean evaluate(Node node)
	{
		boolean ret = true;

		if( node.getAddress() != getDst() )
		{
			return node.routeForNode(this, getDst());
		}
		
		// Look the desired method up on the local node
		CapsuleClass cc = node.lookupCapsuleClass(unknownMID);
		TRACE(Trace.ANTSR_DYNLOAD, "cc for mid " +unknownMID+ " -> " +cc);

		if ((cc == null) || (cc.cl == null))
		{
			/*
			 * This can only happen if the cc is not here
			 * at all or just barely being loaded on this
			 * node.  Whomever asked for it is broken or
			 * trying to annoy me.
			 */

			// XXX these messages should back off exponentially.
			node.log(Log.WARNING,
				 "MIDIDRequest: Ignoring CapsuleClass request from belligerent node "
				 +NodeAddress.toString(getSrc())+ ".  (cc=" +cc+ ")");

			return true;
		}

		// test to see if this is builtin and non-transferable
		if (cc.group.builtin())
		{
			// XXX these messages should back off exponentially.
			node.log(Log.WARNING,
				 "MIDIDRequest: Node " +NodeAddress.toString(getSrc())+
				 " wanted builtin capsule.  Dropping request.");

			return true;
		}
		
		TRACE(Trace.ANTSR_DYNLOAD, "MIDIDRequest for " +unknownMID+ " from " +NodeAddress.toString(getSrc()));
		
		/*
		 * Send off one MIDIDResponseCapsule.
		 */
		Capsule dl = new MIDIDResponseCapsule(getPrevious(),
						      unknownMID,
						      cc.protocolID(),
						      cc.group.groupID,
						      cc.userID()); // XXX

		// send the packet with a TTL of 1 ...
		dl.prime(node.getAddress(), 1);

		return node.routeForNode(dl, getPrevious());
	}
  
	/**
	 * Create a blank MIDIDRequestCapsule.
	 * <p>
	 * This constructor should only be invoked when
	 * a MIDIDRequestCapsule is being received and decoded.
	 */
	PACKAGE MIDIDRequestCapsule()
	{
	}

	/**
	 * Create a MIDIDRequestCapsule that will request info on
	 * the given unknown MID.
	 * <p>
	 * This constructor is only invoked by DLProtocolHandler.  // XXX which is being renamed.
	 */
	PACKAGE MIDIDRequestCapsule(int dst, TypeID unknownMID)
	{
		setDst(dst);
		assert(unknownMID != null);
		this.unknownMID = unknownMID;
	}
	
	public String toString()
	{
		StringBuffer buf = new StringBuffer(100);

		buf.append(super.toString());
		buf.append("+MIDIDRequest[unknownMID=");
		buf.append(unknownMID);
		buf.append("]");
		return buf.toString();
	}
}
