/* 
 *  $Id: TypeID.Java,v 1.1 2000/01/06 01:00:51 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTSR_CORE_PKG;

import JANOS_ANTSR_UTIL_PKG.MD5;

import java.util.Hashtable;
import java.util.Vector;

/**
 * Capsule, CapsuleClass, CodeGroup and Protocol identifiers.
 * This class wraps 128 bit MD5 hashes stored in byte arrays.
 * <p>
 * See the text file <a href="On_IDs.txt">On_IDs.txt</a> for a
 * description of how these IDs are used.
 *
 * @author David Wetherall
 */
public final class TypeID
{
	/**
	 * We only use the first 16 bytes (128 bits)
	 */
	private final byte[] bits;
	
	public int hashCode()
	{
		return ((bits[0] + bits[4] + bits[8] + bits[12]) << 24)
			+ ((bits[1] + bits[5] + bits[9] + bits[13]) << 16)
			+ ((bits[2] + bits[6] + bits[10] + bits[14]) << 8)
			+ (bits[3] + bits[7] + bits[11] + bits[15]);
	}
  
	/**
	 * Compare this TypeID with the given one.  The given
	 * TypeID must match all 16 bytes.
	 */
	public boolean equals(Object obj)
	{
		try
		{
			if (obj == null)
				return false;

			TypeID t = (TypeID)obj;

			for (int i = 0; i < 16; i++)
				if (bits[i] != t.bits[i])
					return false;
      
			return true;
		}
		catch (Exception e)
		{
			return false;
		}
	}

	/**
	 * Table of known TypeIDs.  Used to keep/find canonical
	 * versions of TypeIDs.
	 * <p>
	 * XXX A global table.  Not clear if this really should be
	 * global.
	 */
	private static final Hashtable interned = new Hashtable();


	/**
	 * Intern a TypeID and return its canonical value.
	 *
	 * @return the canonical version of this TypeID.
	 */
	public TypeID intern()
	{
		synchronized(interned)
		{
			TypeID t = (TypeID)interned.get(this);
			if (t != null)
				return t;
			interned.put(this, this);
			return this;
		}
	}

	/**
	 * Compare two TypeIDs
	 * 
	 * @return -1 (this < t), 0 (this.equals(t), 1 (this > t)
	 */
	public int compare(TypeID t)
	{
		for (int i = 0; i < 16; i++)
		{
			if ((bits[i]&0xFF) > (t.bits[i]&0xFF))
				return 1;
			else if ((bits[i]&0xFF) < (t.bits[i]&0xFF))
				return -1;
		}
		return 0;
	}

	/**
	 * Sort an array of TypeIDs.  Good for short vectors
	 * only -- uses insertion sort.  Sorts the array
	 * inline.
	 */
	public static void sortArray(TypeID[] ids)
	{
		for (int j = 1; j < ids.length; j++) {
			TypeID key = ids[j];
			int i = j-1;
			while ((i >= 0)
			       && (ids[i].compare(key) == 1))
			{
				ids[i+1] = ids[i];
				i--;
			}
			ids[i+1] = key;
		}
	}
  
	/**
	 * Compute an ID from a vector of codeIDs.
	 * <p>
	 * This is done by computing a multiple-element md5
	 * hash over a sorted list of code ids.
	 * <p>
	 * XXX Combine the insertion sort and the id.update()
	 * step from this function.  Should only have to
	 * iterate over the list once.
	 *
	 * @return a TypeID representing this vector's ID.
	 */
	public static TypeID computeFromIDs(TypeID[] ids)
	{
		MD5 id = new MD5();

		TypeID.sortArray(ids);
		for (int i = 0; i < ids.length; i++)
		{
			id.update(ids[i].unwrap());
		}

		return TypeID.wrap(id.hash()).intern();
	}

	public String toString()
	{
		if (false)
			return MD5.toString(bits);
		else
			return MD5.toShortString(bits);
	}

	/**
	 * Unwrap this TypeID, returning the inner byte array.
	 * Does not return a copy, returns the actual array.
	 *
	 * @return the byte[] representing this TypeID.
	 */
	public byte[] unwrap()
	{
		return bits;
	}


	/**
	 * Construct and initialize by shallow copy. The different notation
	 * is meant to emphasize that a complete object is not being created
	 * but rather an existing object is being packaged for convenience.
	 * <p>
	 * The length of the buf is not checked.
	 *
	 * @param buf Byte[] containing an MD5 checksum.
	 */ 
	public static TypeID wrap(byte[] buf)
	{
		TypeID t = new TypeID(false, buf, 0);
		return t;
	}

	/**
	 * No-arg constructor is private to prevent construction
	 * of invalid TypeIDs.
	 */
	private TypeID()
	{
		bits = null;
	}

	/**
	 * Construct a new TypeID and initialize by deep copy.
	 *
	 * @param buf the byte[] to copy into this new TypeID.
	 * @param offset offset in buf to copy from
	 * @exception java.lang.IndexOutOfBoundsException if buf is too short
	 */ 
	public TypeID(byte[] buf, int offset)
	{
		// XXX: this(true, buf, offset);  // doesn't see that final 'bits' gets set.
		bits = new byte[16];
		System.arraycopy(buf, offset, bits, 0, 16);
	}
	
	/**
	 * Construct a new type id.  Use the given buffer.
	 * The copybuf argument indicates if the buffer is
	 * copied or simply pointed to.
	 */
	private TypeID(boolean copybuf, byte[] buf, int offset)
	{
		if (copybuf)
		{
			bits = new byte[16];
			System.arraycopy(buf, offset, bits, 0, 16);
		}
		else
		{
			// XXX offset must be 0
			// XXX buf must be the appropriate length
			bits = buf;
		}
	}
}
