/*
 * Log.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_IO_PKG;

import java.util.Vector;
import java.io.ByteArrayOutputStream;
import java.io.PrintWriter;

IMPORT_DEBUG

/**
 * A Janos Log interface.  A log maintains a buffer of a fixed number
 * of messages.  All messages passed to the log via <code>logMsg()</code>
 * are sent to all listeners on the log.
 * <p>
 * Really just a glorified ring buffer.
 * <p>
 * Patterned after syslog(3).
 */
public class Log
{
	/* Log Levels */

	/** System is unusable. */
	public static final int EMERG       = 0; 

	/** Action must be taken immediately. */
	public static final int ALERT       = 1; 

	/** Critical conditions. */
	public static final int CRIT        = 2; 

	/** Error conditions. */
	public static final int ERR         = 3; 

	/** Warning conditions. */
	public static final int WARNING     = 4; 

	/** Normal but significant condition. */
	public static final int NOTICE      = 5; 

	/** Informational messages. */
	public static final int INFO        = 6; 

	/** Debug-level messages. */
	public static final int DBG         = 7; 

	/**
	 * The entities in the ring-buffer.  Each message has its level
	 * recorded along with it.
	 */
	private class Message 
	{
		public int level; // only need 3 bits...
		public String msg;
	}

	/**
	 * The ring buffer.  The msg_ring is never empty.
	 */
	private Message[] msg_ring;

	/**
	 * Size of the ring buffer.
	 */
	private static final int RING_SIZE = 25;
	private int ring_head, ring_tail;

	/**
	 * Listeners on this log.  Should be Vector<LogListener>.
	 */
	private Vector listeners;

	/**
	 * Stores the prefix for the log messages
	 */
	private String prefix;

	private ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
	private PrintWriter byteWriter = new PrintWriter(byteStream);
	
	/**
	 * Create a log.  All messages sent to this log will
	 * be prefixed with the given prefix string.
	 */
	public Log(String prefix)
	{
		int i;
		
		msg_ring = new Message[RING_SIZE];
		for( i = 0; i < RING_SIZE; i++ )
		{
			msg_ring[i] = new Message();
		}
		
		// add a message.  Makes sure log is never empty.
		msg_ring[0].msg = "*** " + prefix + " Log Started";
		msg_ring[0].level = DBG;
		ring_head = 0;
		ring_tail = 0;
		listeners = new Vector();
		this.prefix = prefix;
	}

	/**
	 * Set the prefix message.
	 */
	public void setPrefix(String prefix)
	{
		this.prefix = prefix;
	}
	
	/**
	 * Add a message to this log.  Added to the ring buffer.
	 * All active listeners will recieve the message, too.
	 */
	public void logMsg(int level, String msg)
	{
		// Put the prefix on the message
		if( prefix != null )
			msg = prefix + ":" + msg;

		assert(level >= 0);
		assert(level <= DBG);

		// add to the ring buffer
		ring_tail = (ring_tail + 1) % RING_SIZE;
		msg_ring[ring_tail].level = level;
		msg_ring[ring_tail].msg = msg;
		if (ring_tail == ring_head)
			ring_head = (ring_head + 1) % RING_SIZE;
		
		// pass out to all listeners
		int i;
		for (i = 0; i < listeners.size(); i++)
		{
			LogListener ll;
			ll = ((LogListener) listeners.elementAt(i));
			
			ll.logMsg(level, msg);
		}
	}

	/**
	 * Add a message to the log thats a stack trace
	 */
	public void logThrowable(Throwable th)
	{
		String msg;
		
		if( this.prefix != null )
			this.byteWriter.println(prefix);
		th.printStackTrace(this.byteWriter);
		msg = this.byteStream.toString();
		int i;
		for( i = 0; i < listeners.size(); i++ )
		{
			LogListener ll;
			ll = ((LogListener) listeners.elementAt(i));
			
			ll.logMsg(Log.ERR, msg);
		}
		this.byteStream.reset();
	}
	
	/**
	 * Add a listener to this log.
	 * The listener will immediately receive all buffered messages
	 * and eventually receive all future messages.
	 */
	public void addListener(LogListener ll)
	{
		if (!listeners.contains(ll))
		{
			listeners.addElement(ll);
			int i = ring_head;
			do
			{
				if( msg_ring[i].msg != null )
					ll.logMsg(msg_ring[i].level,
						  msg_ring[i].msg);
				i = (i + 1) % RING_SIZE;
			}
			while (i != ring_tail);
		}
	}

	public void removeListener(LogListener ll)
	{
		listeners.removeElement(ll);
	}
}
