/*
 * StatusMessage.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_IO_PKG;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;

/**
 * A datagram packet that can be used to report information about a node
 */
public class StatusMessage
{
	/**
	 * XXX Max packet size
	 */
	private static final int PACKET_SIZE = 128;

	/**
	 * Method IDs for a packet (i.e. what do they want to know?)
	 */
	public static final byte SMM_NONE = 0;
	public static final byte SMM_QUERY = 1;

	/**
	 * Flags for the message
	 */
	public static final byte SMF_REPLY = 1; /* Message is a reply */

	/**
	 * Byte array for the packet data
	 */
	private byte data[] = new byte[PACKET_SIZE];

	/**
	 * The real packet that gets sent back and forth
	 */
	private DatagramPacket dp = new DatagramPacket(data, PACKET_SIZE);

	/**
	 * Encode a short in the byte array at the given offset
	 */
	public static void encodeShort(byte data[], int offset, short s)
	{
		data[offset] =   (byte)((s >> 8) & 0xFF);
		data[offset+1] = (byte) (s       & 0xFF);
	}
	
	/**
	 * Encode an int in the byte array at the given offset
	 */
	public static void encodeInt(byte data[], int offset, int i)
	{
		data[offset] =   (byte)((i >> 24) & 0xFF);
		data[offset+1] = (byte)((i >> 16) & 0xFF);
		data[offset+2] = (byte)((i >>  8) & 0xFF);
		data[offset+3] = (byte) (i        & 0xFF);
	}
	
	/**
	 * decode a short in the byte array at the given offset
	 */
	public static short decodeShort(byte data[], int offset)
	{
		short retval = 0;

		retval |= data[offset] << 8;
		retval |= data[offset+1];
		return retval;
	}
	
	/**
	 * decode an int in the byte array at the given offset
	 */
	public static int decodeInt(byte data[], int offset)
	{
		int retval = 0;

		retval |= data[offset] << 24;
		retval |= data[offset+1] << 16;
		retval |= data[offset+2] << 8;
		retval |= data[offset+3];
		return retval;
	}

	/**
	 * Get the real datagram packet
	 */
	public DatagramPacket getDatagramPacket()
	{
		return this.dp;
	}

	/**
	 * Set the peer address
	 */
	public void setAddress(InetAddress ia)
	{
		this.dp.setAddress(ia);
	}

	/**
	 * Get the peer address
	 */
	public InetAddress getAddress()
	{
		return this.dp.getAddress();
	}

	/**
	 * Set the peer port
	 */
	public void setPort(int port)
	{
		this.dp.setPort(port);
	}
	
	/**
	 * Get the peer port
	 */
	public int getPort()
	{
		return this.dp.getPort();
	}

	/**
	 * Set the method for this message
	 */
	public void setMethodID(byte method)
	{
		data[0] = method;
	}

	/**
	 * Get the method for this message
	 */
	public byte getMethodID()
	{
		return data[0];
	}

	/**
	 * Set the flags for this message
	 */
	public void setFlags(byte flags)
	{
		data[1] = flags;
	}

	/**
	 * Get the flags for this message
	 */
	public byte getFlags()
	{
		return data[1];
	}

	/**
	 * Set the size of the message payload
	 */
	public void setValueSize(int size)
	{
		encodeInt(data, 4, size);
	}

	/**
	 * Get the size of the message payload
	 */
	public int getValueSize()
	{
		return decodeInt(data, 4);
	}

	/**
	 * Get the byte array for the packet
	 */
	public byte []getValue()
	{
		return data;
	}

	/**
	 * Get the offset into the value array where the value actually starts
	 */
	public int getValueOffset()
	{
		return 8;
	}
}
