/*
 * CommSpace.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.lang.ArrayIndexOutOfBoundsException;

/**
 * Variation on an array, with some tuple-space like properties.
 * Objects are not shared between flows, they are copied on read.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class CommSpace
{
	private final CommSpaceElement[] table;
	private int currentSize; // XXX not used..
	
	PACKAGE CommSpace(int size)
	{
		this.table = new CommSpaceElement[size];
	}

	PACKAGE final int size()
	{
		return table.length;
	}

	private final void checkIndex(int index)
	{
		if ((index > table.length)
		    || (index < 0))
			throw new ArrayIndexOutOfBoundsException("table only has "
								 +table.length+ " elements");
	}

	/**
	 * Write the given element at the given index.  Note
	 * that this creates cross-process pointers..
	 */
	PACKAGE void write(int index, CommSpaceElement elem)
		throws ArrayIndexOutOfBoundsException
	{
		checkIndex(index);
		synchronized(table)
		{
			table[index] = elem;
			table.notifyAll(); // XXX GAK!
		}
	}

	PACKAGE CommSpaceElement take(int index)
		throws ArrayIndexOutOfBoundsException
	{
		CommSpaceElement e;

		checkIndex(index);
		synchronized(table)
		{
			e = table[index];
			table[index] = null;
		}
		return e;
	}

	PACKAGE CommSpaceElement waitToTake(int index)
		throws ArrayIndexOutOfBoundsException, InterruptedException
	{
		CommSpaceElement e;

		checkIndex(index);
		synchronized(table)
		{
			e = table[index];
			while (e == null)
			{
				table.wait(); // XXX GAK!
				e = table[index];
			}
			table[index] = null;
		}
		return e;
	}

	PACKAGE CommSpaceElement read(int index)
		throws ArrayIndexOutOfBoundsException
	{
		CommSpaceElement e;

		checkIndex(index);
		synchronized(table)
		{
			e = table[index];
		}
		return e;
	}

	PACKAGE CommSpaceElement swap(int index, CommSpaceElement elem)
	{
		CommSpaceElement e;
		
		checkIndex(index);
		synchronized(table)
		{
			e = table[index];
			table[index] = elem;
			table.notifyAll(); // XXX GAK!
		}
		return e;
	}
}
