/*
 * Flow.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX

/**
 * Java wrapper for a flow.
 * A flow is the closest approximation to a process in Janos.
 * <p>
 * XXX: A lot of this will have to change so that starting a Flow
 * starts a new process (somehow).
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class Flow
{
	/**
	 * The credentials provided when this Flow was created.  This
	 * is the bucket from which the Flow gets to extract resources
	 * and access Node facilities.
	 */
	private final Credentials cred;

	/**
	 * A ThreadPool to be used for Flow-specific actions, mostly
	 * just initialization and termination handling.
	 */
	private final ThreadPool tPool;

	/**
	 * The Memory limit for this flow.
	 */
	private final MemSpec memSpec;

	/**
	 * Channels created in this Flow.  This list only exists
	 * so that the channels can be cleanly destroyed when the
	 * flow is terminated.
	 */
	private final Vector channels = new Vector(5);

	/**
	 * The parent flow of this flow.  Used for handling termination
	 * requests.
	 * <p>
	 * XXX this is a cross-process pointer and must be treated
	 * specially.
	 */
	private final Flow parentFlow;

	/**
	 * A per-flow Object that user code can set and get.
	 * (for associating things with the flow).
	 */
	private Object obj;

	boolean destroyed = false;
	
	/**
	 * The Flow ClassLoader for this flow.
	 * <p>
	 * This object is allocated and initialized by the flow's run()
	 * method.  Thus it isn't final.
	 *
	 * @see FlowInitializer#run()
	 */
	PACKAGE FlowClassLoader loader;

	/**
	 * Per-flow data.
	 */
	PACKAGE final FlowState state;

	PACKAGE FlowRun flowRun;

	/**
	 * Create a new Flow, which is a child of the current Flow.
	 * <p>
	 * This is the only interface for creating a Flow.
	 * <p>
	 * The termHandler will be invoked by a thread in the parent
	 * Flow's if the child Flow is terminated.  This handler has a
	 * short limit for execution.
	 * <p>
	 * If termHandler is null, then no termination handler is called.
	 * <p>
	 * XXX Should be called "createFlow()" to sync back with NodeOS spec.
	 *
	 * @param cred The principal associated with this Flow
	 * @param tPoolSpec The specification for the new Flow's ThreadPool
	 * @param memSpec The memory limit for this flow
 	 * @param flowStartSpec Class loading and startup information
	 * @param termHandler Termination handler invoked by the NodeOS when an immediate child
	 * flow is being terminated.
	 */
	public FlowHandle newFlow(final Credentials cred,
				  final ThreadPoolSpec tPoolSpec,
				  final MemSpec memSpec,
				  final FlowStartSpec flowStartSpec,
				  final Runnable termHandler)
	{
		Flow flow = new Flow(this,
				     cred,
				     tPoolSpec,
				     memSpec,
				     flowStartSpec);
		return new FlowHandle(flow, termHandler);
	}

	/**
	 * Initalize a new Flow.  See newFlow().
	 * <p>
	 * XXX implicitly starts threads.  Gak.
	 * 
	 * @allocation: allocates a FlowState() and ThreadPool (in the
	 * parent's context).
	 */
	PACKAGE Flow(final Flow parent,
		     final Credentials cred,
		     final ThreadPoolSpec tPoolSpec,
		     final MemSpec memSpec,
		     final FlowStartSpec flowStartSpec)
	{
		this.parentFlow = parent;
		this.cred = cred;
		this.memSpec = memSpec;

		this.state = new FlowState(memSpec);
		this.state.init(this);

		this.tPool = new ThreadPool(this,
			tPoolSpec.variousSchedulerParameters,
			tPoolSpec.scheduler);

		FlowRegister.register(this); // XXX force a name?

		/*
		 * Create a FlowInitializer to finish the Flow
		 * initialization in the context of a thread in the
		 * new ThreadPool.
		 */
		FlowInitializer fi;
		fi = new FlowInitializer(this, tPool, flowStartSpec);
		tPool.threadStart(fi).setName("Flow init "
					      + flowStartSpec.startClass);
	}

	public synchronized void destroy()
	{
		if( this.destroyed )
			return;
		this.destroyed = true;
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);

		// parentFlow.destroyCallback();

		for (int i = 0; i < channels.size(); i++)
		{
			// Yes, this is stupid.  It just temporary.

			Object c = channels.elementAt(i);
			if (c instanceof InChannel)
			{
				((InChannel)c).destroy();
			}
			else if (c instanceof OutChannel)
			{
				((OutChannel)c).destroy();
			}
			else if (c instanceof CutThroughChannel)
			{
				((CutThroughChannel)c).destroy();
			}
			else
			{
				panic("Non-channel object in channel list.");
			}
		}

		tPool.destroy();

		loader.destroy();

		FlowRegister.unregister(this);
	}
	
	PACKAGE void addChannel(InChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE void addChannel(OutChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE void addChannel(CutThroughChannel c)
	{
		channels.addElement(c);
	}

	public Object getObject()
	{
		return obj;
	}

	public void setObject(Object obj)
	{
		this.obj = obj;
	}

	public ThreadPool threadPool()
	{
		return tPool;
	}

	public String toString()
	{
		if (false) // longwinded
			return "Flow["
				DEBUGONLY(+ "name='" +name)
				+ "; parent=" + ((parentFlow==null)?"no":"yes")
				+ "; loader=" + loader
				+ "; cred=" + cred
				+ "; obj=" + obj
				+ "]";
		else // short and sweet
			return "Flow[" 
				DEBUGONLY(+ "name='" +name)
				+ "; parent=" + ((parentFlow==null)?"n":"y")
				+ "; loader=" + loader
				+ "]";
	}

	/**
	 * Add a new class to an existing flow.
	 */
	public void addClass(String name, byte[] bytecode)
	{
		assert(this == Flow.currentFlow());
		assert(name != null);
		assert(bytecode != null);
		assert(loader);
		loader.addClass(name, bytecode);
	}
	
	/**
	 * Add a package to the list of "system" packages.  System
	 * packages will be loaded by the system class loader.
	 * All other classes will be loaded by the per-flow
	 * classloader.
	 * <p>
	 * All sub-flows *share* the same definition of "system"
	 * packages for now.  This is because only one EE System
	 * is loaded on the NodeOS.  This will have to change.
	 * <p>
	 * Implicitly allows the package.
	 */
	public void addSystemPackage(String name)
	{
		assert(this == Flow.currentFlow());
		assert(name != null);
		assert(loader != null);

		loader.addSystemPackage(name);
	}
	 
	/**
	 * Allow access to a Java package.  Must be invoked in the
	 * context of this flow.
	 * <p>
	 * addLocalPackage implicitly allows those packages, so
	 * addPackage need not be called on those packages.
	 * <p>
	 * Allowed packages that are not local are either dynamically
	 * loaded over the net, or are found locally.
	 */
	public void allowPackage(String pkgPrefix)
	{
		assert(this == Flow.currentFlow());
		assert(pkgPrefix != null);
		assert(loader != null);

		loader.allowPackage(pkgPrefix);
	}
	
	public void enablePackageControls(boolean doit)
	{
		assert(this == Flow.currentFlow());
		assert(loader != null);

		loader.enablePackageControls(doit);
	}

	public Class loadClass(String name)
		throws java.lang.ClassNotFoundException
	{
		assert(name != null);
		assert(loader);
		return loader.loadClass(name);
	}

	/**
	 * Get the current flow.  The current Flow is the
	 * flow associated with the current Thread.
	 *
	 * @see Thread#getFlow()
	 */
	public final static Flow currentFlow()
	{
		return Thread.self().getFlow();
	}

#ifdef DEBUG
	// always non-null.
	private String name = "John Doe Flow";
	
	public void setName(String name)
	{
		if (name != null)
			this.name = name;
	}

	public String getName()
	{
		return name;
	}
#endif /*DEBUG*/	
}

