/*
 * InChannel.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX
import java.io.IOException;

/**
 * Java wrapper for an InChannel.
 *
 * @see CutThroughChannel
 * @see OutChannel
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class InChannel
{
	/**
	 * Queue of free packet buffers for this InChannel.
	 * Queue is circular, doubly-linked list.  freeBufferQ points
	 * to the "head", freeBufferQ.prev points to the tail.
	 */
	private ProtocolLayer.PacketBuffer freeBufferQ;

	/**
	 * Lock for protecting the free packet buffer queue.
	 */
	private final Object freeBufferQLock = new Object(); // XXX

	/**
	 * Dispatch an incoming packet for this in channel.  If the channel can
	 * afford the packet one of its own is returned as an exchange.  If the
	 * channel can't afford the packet then the original one is
	 * returned.
	 */
	PACKAGE void dispatchBuffer(ProtocolLayer.PacketBuffer pb,
				    int packetStart,
				    int payloadStart)
	{
		PacketWorker worker = PacketWorker.getFree(flow);
		
		worker.deliverFunc = this.deliverFunc;
		pb.acquire();
		worker.bufferHandle.setBuffer(pb, packetStart, pb.length);
		worker.payloadStart = payloadStart;
		threadPool.addWork(worker.workItem);
	}

	/**
	 * Add a buffer to the InChannel
	 */
	public void addBuffer(ProtocolLayer.PacketBuffer pb)
	{
		synchronized(this.freeBufferQLock)
		{
			freeBufferQ = pb.addToQueue(freeBufferQ);			
		}
	}

	/**
	 * Remove a buffer from the queue of free buffers
	 * on this InChannel.  Returns null if there are
	 * no free buffers on this InChannel.
	 */
	public ProtocolLayer.PacketBuffer remBuffer()
	{
		ProtocolLayer.PacketBuffer retval = null;

		synchronized(this.freeBufferQLock)
		{
			if (freeBufferQ != null)
			{
				retval = freeBufferQ;
				freeBufferQ = freeBufferQ.removeFromQueue();
			}
		}
		return retval;
	}
	
	/**
	 * Add `count' buffers to the in channel.
	 * <p>
	 * XXX also adds `count' workers to the associated flow.
	 */
	public void addBuffers(int count)
	{
		int lpc;
		
		require(count > 0);

		for( lpc = 0; lpc < count; lpc++ )
		{
			this.addBuffer(new ProtocolLayer.PacketBuffer(this));
		}
		PacketWorker.addWorkers(this.flow, count);
	}

	/**
	 * Remove `count' buffers from the channel.  Or all the buffers
	 * if there are less than `count'.
	 * <p>
	 * XXX does not remove `count' workers from the associated flow...
	 */
	public void remBuffers(int count)
	{
		int lpc;
		
		require(count > 0);

		for( lpc = 0; lpc < count; lpc++ )
		{
			Object o;
			o = this.remBuffer();
			if (o == null)
				return;
		}
	}
	
	/**
	 * Death Packet. XXX.
	 */
	public static final byte[] exitPacket = {-0x22, -0x53, -0x42, -0x11};

	/**
	 * The flow this receive channel is associated with.
	 */
	PACKAGE final Flow flow;

	/**
	 * The demultiplex key associated with this channel.
	 */
	PACKAGE final DemultiplexKey classifier;

	/**
	 * The ThreadPool to send incoming packets to.
	 * ThreadPool.addWork() had best be protected kernel code
	 * (i.e., final).
	 */
	private final ThreadPool threadPool;

	private InChannelDispatcher handler;
	
	/**
	 * The callback for delivery.
	 */
	private final DeliverFunc deliverFunc;

	public InChannel(Flow flow,
			 DemultiplexKey classifier,
			 String protocolSpec,
			 String addrSpec,
			 ThreadPool threadPool,
			 DeliverFunc deliverFunc)
		throws IOException,
		       ProtocolLayer.PacketAddressFormatException,
		       ProtocolSpec.FormatException
	{
		ProtocolLayer pl;
		ProtocolSpec ps;
		
		this.flow = flow;
		this.classifier = classifier;
		this.threadPool = threadPool;
		this.deliverFunc = deliverFunc;

		assert(classifier);
		// XXX doesn't sufficiently protect the kernel...
		if((flow == null)
		   || (threadPool == null)
		   || (classifier == null)
		   || (protocolSpec == null)
		   || (threadPool == null)
		   || (deliverFunc == null))
			throw new NullPointerException();

		/* Find the matching protocol layer */
		ps = new ProtocolSpec(protocolSpec);
		if( (pl = ProtocolLayer.
		     findLayer(ps.getComponent(ps.componentCount() - 1)))
		    != null )
		{
			ProtocolLayer.PacketAddressMask source;
			InChannelDispatcher icd;

			/* Get the source address mask */
			source = pl.createAddressMask();
			source.parseAddressMask(addrSpec);
			icd = InChannelDispatcher.getDispatcher(pl, source);
			this.handler = icd;
			icd.addChannel(this);
		}
		flow.addChannel(this);
	}
	
	public String toString()
	{
		return "InChannel[flow="
			+ flow
			+ "; handler="
			+ handler
			+ "; classifier="
			+ classifier
			+ "; threadPool="
			+ threadPool
			+ "; deliverFunc="
			+ deliverFunc
			+ "]";
	}

	PACKAGE void destroy()
	{
		this.handler.remChannel(this);
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
	}
	
	protected void finalize()
	{
		this.handler.remChannel(this);
	}
}
