/*
 * InChannel.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;
import java.util.Vector;
import java.io.IOException;

/**
 * Class that is used to dispatch packets from protocol layers to in channels
 */
PACKAGE class InChannelDispatcher
	extends ProtocolLayer.PacketHandler
{
	/**
	 * The set of dispatchers for the system.
	 */
	private static final Hashtable dispatchers = new Hashtable();

	/**
	 * Get the dispatcher that matches the given source mask.
	 *
	 * @param pl The protocol layer were packets are coming from
	 * @param source The source address mask packets are matched too
	 */
	PACKAGE static InChannelDispatcher getDispatcher(
		ProtocolLayer pl, ProtocolLayer.PacketAddressMask source)
		throws IOException
	{
		InChannelDispatcher retval;

		synchronized(InChannelDispatcher.dispatchers)
		{
			/*
			 * Search our hash table first, and then mask one if
			 * needed
			 */
			if( (retval = (InChannelDispatcher)InChannelDispatcher.
			     dispatchers.get(source)) == null )
			{
				retval = new InChannelDispatcher();
				retval.protocolLayer = pl;
				retval.source = source;
				InChannelDispatcher.dispatchers.put(source,
								    retval);
				pl.attachHandler(retval);
			}
		}
		return retval;
	}

	/**
	 * The protocol layer that this dispatcher is attached to
	 */
	private ProtocolLayer protocolLayer;
	
	/**
	 * The set of channels that this handler can dispatch too.
	 * We only really care about their demultiplex keys since
	 * the rest of the demultiplex is done elsewhere.
	 */
	private final Vector channels = new Vector();

	/**
	 * Add a channel to this dispatcher.  The channel is placed
	 * in the selection line by how large its key is, the larger
	 * the key the earlier it will be checked.
	 */
	PACKAGE void addChannel(InChannel newChan)
	{
		int lpc, size;

		synchronized(this.channels)
		{
			size = this.channels.size();
			for( lpc = 0; lpc < size; lpc++ )
			{
				InChannel elem = (InChannel)this.channels.
					elementAt(lpc);

				/* Insert in biggest to smallest order */
				if( newChan.classifier.length() >=
				    elem.classifier.length() )
				{
					break;
				}
			}
			TRACE(Trace.NODEOS_DISPATCH, "Adding packet classifier " + newChan.classifier);
			this.channels.insertElementAt(newChan, lpc);
		}
	}

	/**
	 * Remove a channel from this dispatcher.
	 */
	PACKAGE void remChannel(InChannel deadChan)
	{
		synchronized(this.channels)
		{
			this.channels.removeElement(deadChan);
			if( this.channels.size() == 0 )
			{
				/*
				 * No more channels, detach ourselves from the
				 * protocol layer
				 */
				this.protocolLayer.detachHandler(this);
				synchronized(InChannelDispatcher.dispatchers)
				{
					InChannelDispatcher.dispatchers.
						remove(this.source);
				}
			}
		}
	}

	/**
	 * Handle a packet that came in through the transport layer.
	 */
	public ProtocolLayer.PacketBuffer handle(ProtocolLayer.PacketBuffer pb)
	{
		ProtocolLayer.PacketBuffer retval = pb;
		InChannel matchChan = null;
		int payloadStart = 0;
		int lpc, size;

		/**
		 * Walk all the channels attached to this handler
		 * and find a match.
		 */
		synchronized(this.channels)
		{
			size = this.channels.size();
			for( lpc = 0;
			     (lpc < size) && (matchChan == null);
			     lpc++ )
			{
				InChannel elem = (InChannel)this.channels.
					elementAt(lpc);

				payloadStart = elem.classifier.
					matches(pb.data, 0, pb.length);
				if( payloadStart != 0 )
				{
					/* Found a match */
					matchChan = elem;
				}
			}
		}
		if( matchChan != null )
		{
			TRACE(Trace.NODEOS_INCOMING, "demultiplexed to: " +matchChan.classifier);
			/*
			 * We need to return a buffer, so try and get one from
			 * the channel
			 */
			if( (retval = matchChan.remBuffer()) != null )
			{
				retval.trade(pb);
				matchChan.dispatchBuffer(pb, 0, payloadStart);
			}
			else
			{
				retval = pb;
				/*
				 * The channel can't afford the packet,
				 * just return the passed in packet
				 */
				TRACE(Trace.NODEOS_INCOMING, "Channel can't afford packet.");
			}
		}
		else
		{
			TRACE(Trace.NODEOS_INCOMING, "Can't schedule pkt: data=" +Debug.dbgPacketSummary(pb));
		}
		return retval;
	}
}
