/*
 * NativeThread.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

/**
 * A "Native" Thread.
 * <p>
 * A package-private wrapper for the "native" thread
 * layer.
 * <p>
 * This is a simple wrapper for java.lang.Thread.  Its
 * mostly a separate class so I can cleanly subclass it
 * without exposing all of the standard java.lang.Thread
 * methods.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class NativeThread
	extends java.lang.Thread
{
	/**
	 * InfoHook class that reports on all of the native threads in the
	 * system.
	 */
	private static class NativeThreadInfo
		extends InfoHook
	{
		private NativeThreadInfo()
		{
			super("threads");
		}
		
		public String[] getInfo()
		{
			String[] retval;
			java.lang.Thread[] allThreads;
			ThreadGroup r, p;

			// find root ThreadGroup
			r = NativeThread.currentThread().getThreadGroup();
			p = r;
			while (p != null)
			{
				r = p;
				p = r.getParent();
			}
			assert(r != null);

			allThreads = new java.lang.Thread[r.activeCount() + 50];
			int ct = r.enumerate(allThreads, true);
			
			retval = new String[ct];
			for (int i = 0; i < ct; i++)
			{
				retval[i] = allThreads[i].toString()
					+ " name="
					+ allThreads[i].getName();
			}

			return retval;
		}
	}

	static
	{
		Status.addInfoHook(new NativeThreadInfo());
	}


	/**
	 * The Janos Thread that wraps this native thread.
	 */
	PACKAGE final Thread janosThread;

	/**
	 * The Runnable to use; otherwise defaults to the
	 * run method on this (which just invokes the JanosThread.run()).
	 * <p>
	 * @todo Make this non-final so it can be nulled out?
	 */
	private final java.lang.Runnable r;

	/**
	 * Create a new native thread.  This constructor is
	 * used to only create "implicit" threads.  Implicit
	 * threads are marked as daemons.	 
	 * <p>
	 * Only called by Thread constructor.
	 *
	 * @allocations: None (modulo super.<init>())
	 * @see Thread#Thread(ThreadPool)
	 */
	PACKAGE NativeThread(Thread janosThread)
	{
		r = null;
		this.janosThread = janosThread;

		/*
		 * In debug versions, set the default priority
		 * to 5.  That way, code that prints fatal
		 * messages can be bumped up before printing.
		 *
		 * XXX this hack works until we pretend to do thread scheduling.
		 */
		DEBUGONLY(setPriority(5););
		setDaemon(true);
	}

	/**
	 * Create a new native thread.  This constructor is
	 * used to only create "explicit" threads.  Explicit
	 * threads are not daemons.
	 * <p>
	 * The given Runnable will be used when the
	 * thread is started (instead of the run method on this).
	 *
	 * @allocations: None (modulo super.<init>())
	 */
	PACKAGE NativeThread(Thread janosThread, Runnable r)
	{
		this.r = r;
		this.janosThread = janosThread;

		DEBUGONLY(setPriority(5););
		setDaemon(false);
	}

	PACKAGE static void yieldTo(NativeThread th)
	{
		yield();
	}

	public String toString()
	{
		return "Janos NativeThread@0x" + Integer.toHexString(this.hashCode());
	}

	public void run()
	{
		try
		{
			if (r != null)
				r.run();
			else
				janosThread.run();
		}
		catch(Throwable t)
		{
			System.out.println("Exception in thread " +this+ " (runnable = " +r+ ")");
			System.out.println("Thread blew up with:" +t.getMessage());
			t.printStackTrace();
		}
	}
}
