/*
 * OutChannel.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;

import java.net.DatagramSocket;
import java.net.DatagramPacket;

/**
 * Java wrapper for a Channel.
 *
 * @see CutThroughChannel
 * @see InChannel
 */
public final class OutChannel
	extends ProtocolLayer.PacketGenerator
{
	/**
	 * Queue of buffers to be sent out.
	 */
	private final Object bufferQLock = new Object(); // XXX optimize this
	private ProtocolLayer.PacketBuffer bufferQ;

	PACKAGE final Flow flow;
	PACKAGE final OffsetLength ol = new OffsetLength();
	PACKAGE ProtocolLayer protocolLayer;

	public OutChannel(Flow flow,
			  String protocolSpec,
			  String addressSpec)
		throws IOException,
		       ProtocolLayer.PacketAddressFormatException,
		       ProtocolSpec.FormatException
	{
		ProtocolLayer.PacketAddress pa;
		ProtocolLayer pl;
		ProtocolSpec ps;
		
		if( (flow == null) ||
		    (protocolSpec == null) ||
		    (addressSpec == null) )
			throw new NullPointerException();

		this.flow = flow;
		/* Find the matching protocol layer */
		ps = new ProtocolSpec(protocolSpec);
		if( (pl = ProtocolLayer.
		     findLayer(ps.getComponent(ps.componentCount() - 1)))
		    != null )
		{
			ProtocolLayer.PacketAddress dest;

			/*
			 * Get an address object corresponding to the
			 * destination
			 */
			dest = pl.createAddress();
			dest.parseAddress(addressSpec);
			this.destination = dest;
			pl.attachGenerator(this);
			this.protocolLayer = pl;
		}
		else
		{
			throw new IOException();
		}
		flow.addChannel(this);
	}

	/**
	 * Packet generator function defined by ProtocolLayer.PacketGenerator.
	 */
	public ProtocolLayer.PacketBuffer generate()
	{
		ProtocolLayer.PacketBuffer retval = null;
		
		synchronized(this.bufferQLock)
		{
			if (this.bufferQ != null) 
			{
				retval = this.bufferQ;
				bufferQ = bufferQ.removeFromQueue();
			}
		}
		return retval;
	}

	/**
	 * Queue a packet for sending
	 */
	public void queuePacket(ProtocolLayer.PacketBuffer pb)
	{
		synchronized(this.bufferQLock)
		{
			/*
			 * Need to add to the "end" of the queue,
			 * so add it as if on bufferQ.prev was head.
			 *
			 * Empty or singleton lists, add it to the "front".
			 */
			if ((bufferQ == null) || (bufferQ.next == bufferQ))
				bufferQ = pb.addToQueue(bufferQ);
			else
				pb.addToQueue(bufferQ.prev);
		}
	}

	/**
	 * Send a single buffer handle.
	 */
	public void send(BufferHandle bh)
	{
		ProtocolLayer.PacketBuffer pb;

		pb = bh.getBuffer(this.ol);
		if( this.ol.offset == 0 )
		{
			pb.length = this.ol.length;
		}
		else
		{
			ProtocolLayer.PacketBuffer newBuffer;

			newBuffer = new ProtocolLayer.PacketBuffer(null);
			newBuffer.acquire();
			System.arraycopy(pb.data, ol.offset,
					 newBuffer.data, 0,
					 ol.length);
			newBuffer.length = ol.length;
			pb.release();
			pb = newBuffer;
		}
		TRACE(Trace.NODEOS_OUTGOING, "sending: destination=" +this+ " data=" +Debug.dbgPacketSummary(pb));
		this.queuePacket(pb);
		this.protocolLayer.generatorReady(this);
		pb.release();
	}
	
	/**
	 * Send an array of buffer handles.
	 */
	public void sendv(BufferHandle bh)
	{
		if( bh.getNext() == bh )
		{
			this.send(bh);
		}
		else
		{
			ProtocolLayer.PacketBuffer pb;
			int totalLength;
			
			totalLength = BufferHandle.aggregateLength(bh);
			pb = new ProtocolLayer.PacketBuffer(null);
			BufferHandle.aggregate(pb.data, bh);
			TRACE(Trace.NODEOS_OUTGOING, "sending: destination=" +this+ " data=" +Debug.dbgPacketSummary(pb));
			this.queuePacket(pb);
			this.protocolLayer.generatorReady(this);
		}
	}

	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		this.protocolLayer.detachGenerator(this);
	}

	public String toString()
	{
		return "OutChannel:"
			+ super.toString()
			+ "[flow="
			+ this.flow
			+ "; protocolLayer="
			+ this.protocolLayer
			+ "]";
	}
}
