/*
 * ProtocolLayer.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;
import java.io.IOException;

/**
 * ProtocolLayer is supposed to act as an abstraction for simulating
 * transport layers.  Its made up of several interfaces and classes for
 * moving packets through the layer to their destination.
 */
public abstract class ProtocolLayer
{
	/**
	 * The maximum packet length for all the layers.  Since packets are
	 * exchanged with users of the layers we have to make packets big
	 * enough to accomodate all the different protocol layers.
	 */
	private static int maxPacketLength = 0;

	/**
	 * Return the current value of maxPacketLength.
	 */
	PACKAGE static int getMaxPacketLength()
	{
		return ProtocolLayer.maxPacketLength;
	}

	/**
	 * The table of all registered protocol layers.
	 */
	private static Hashtable layers = new Hashtable();

	/**
	 * Install a new protocol layer, basically, this just makes the layer
	 * PACKAGEally available.
	 */
	PACKAGE static void installLayer(ProtocolLayer pl)
	{
		ProtocolLayer.layers.put(pl.getID(), pl);
		ProtocolLayer.maxPacketLength =
			Math.max(ProtocolLayer.maxPacketLength,
				 pl.getPacketLength());
	}

	/**
	 * Find an installed protocol layer with the given name.
	 */
	PACKAGE static ProtocolLayer findLayer(String id)
	{
		return (ProtocolLayer)ProtocolLayer.layers.get(id);
	}

	/**
	 * The global packet recycler.
	 */
	private static PacketRecycler recycler;
	
	PACKAGE static void attachRecycler(PacketRecycler pr)
	{
		ProtocolLayer.recycler = pr;
	}

	PACKAGE static void recycle(PacketBuffer pb)
	{
		ProtocolLayer.recycler.recycle(pb);
	}

	PACKAGE static void detachRecycler(PacketRecycler pr)
	{
		ProtocolLayer.recycler = null;
	}

	/**
	 * Return the string identifier for this protocol layer.
	 */
	PACKAGE abstract String getID();
	/**
	 * Get the maximum packet length for this protocol layer.  This
	 * is pretty much for internal use only.
	 */
	PACKAGE abstract int getPacketLength();
	/**
	 * Create an address object that is compatible with this protocol
	 * layer.
	 */
	PACKAGE abstract PacketAddress createAddress();
	/**
	 * Create an address mask object that is compatible with this
	 * protocol layer.
	 */
	PACKAGE abstract PacketAddressMask createAddressMask();
	/**
	 * Attach a packet handler to the layer.  The source mask for the
	 * handler determines what packets will be dispatched to the handler.
	 */
	PACKAGE abstract void attachHandler(PacketHandler ph)
		throws IOException;
	/**
	 * Detach a packet handler from the layer.
	 */
	PACKAGE abstract void detachHandler(PacketHandler ph);
	/**
	 * Attach a packet generator to the layer.  The destination of the
	 * packet is determined by the destination address object in the
	 * packet generator object.
	 */
	PACKAGE abstract void attachGenerator(PacketGenerator pg);
	/**
	 * Signal that a generator has packets and is ready to be serviced.
	 */
	PACKAGE abstract void generatorReady(PacketGenerator pg);
	/**
	 * Detach a packet generator from the protocol layer.
	 */
	PACKAGE abstract void detachGenerator(PacketGenerator pg);

	/**
	 * A simple exception class for signalling a format error in an
	 * address string.
	 */
	public static class PacketAddressFormatException
		extends Exception
	{
	}

	/**
	 * An interface that represents a host address in the format for
	 * the protocol layer.
	 */
	PACKAGE interface PacketAddress
	{
		/**
		 * Parses a specially formatted string into an internal
		 * structure.  The format of the string is specific to
		 * the protocol layer.
		 */
		void parseAddress(String addr)
			throws PacketAddressFormatException;
	}

	/**
	 * An interface that represents a mask that can be applied to
	 * incoming packets to determine matching packets.
	 */
	PACKAGE interface PacketAddressMask
	{
		/**
		 * Identical to PacketAddress.parseAddress().
		 */
		void parseAddressMask(String addr)
			throws PacketAddressFormatException;
	}

	/**
	 * Object used to point to the actual data buffers for packets.
	 */
	PACKAGE static final class PacketBuffer
	{
		PACKAGE final byte[] data;
		PACKAGE Object owner;
		PACKAGE int length;
		PACKAGE int users;

		/**
		 * For the free buffer queue on InChannels and OutChannels.
		 *
		 * @see InChannel
		 * @see OutChannel
		 */
		PACKAGE PacketBuffer next;

		/**
		 * For the free buffer queue on InChannels and OutChannels.
		 *
		 * @see InChannel
		 * @see OutChannel
		 */
		PACKAGE PacketBuffer prev;

		/**
		 * @allocation: allocates a new max-length byte array.
		 */
		PACKAGE PacketBuffer(Object owner)
		{
			this(owner,
			     new byte[ProtocolLayer.getMaxPacketLength()]);
		}
		
		PACKAGE PacketBuffer(Object owner, byte[] data)
		{
			this.owner = owner;
			this.data = data;
			this.length = 0;
			this.users = 0;
		}

		/**
		 * Debugging only.  Returns true if this buffer is
		 * not on a queue (i.e, next == prev == null).
		 */
		PACKAGE boolean notOnAQueue()
		{
			// Wholly on or, wholly off.
			assert(((next == null) && (prev == null))
			       || ((next != null) && (prev != null)));
			// Singleton list, or not
			assert(((next == this) && (prev == this))
			       || ((next != this) && (prev != this)));

			return ((next == null) && (prev == null));
		}

		/**
		 * Debugging only.  Returns true if this buffer is on
		 * a queue.
		 */
		PACKAGE boolean onAQueue()
		{
			// Wholly on or, wholly off.
			assert(((next == null) && (prev == null))
			       || ((next != null) && (prev != null)));
			// Singleton list, or not
			assert(((next == this) && (prev == this))
			       || ((next != this) && (prev != this)));

			return ((next != null) && (prev != null));
		}
 
		/**
		 * Add this packet buffer to the "head" of given queue q.
		 * Q is doubly linked, the new head (this) is returned.
		 *
		 * @param q pointer to the first element of the list (or null).
		 * @return the new head of the list (always this).
		 */
		PACKAGE PacketBuffer addToQueue(PacketBuffer q)
		{
			require(this.notOnAQueue());

			if (q == null)
			{
				this.next = this;
				this.prev = this;
			}
			else
			{
				assert(q != this);
				assert(q.next != this);
				assert(q.prev != this);
			
				assert(q.prev != null);
				require(q.onAQueue());
				require(q.prev.onAQueue());

				this.next = q;
				this.prev = q.prev;

				// fixup current first and last
				q.prev.next = this;
				q.prev = this;
			}
			
			ensure(this.onAQueue());

			// return new head of q -- this
			return this;
		}
	
		
		/**
		 * Remove this pb from the queue it is on.  It
		 * must be on a queue.
		 *
		 * @return the new <emph>head</emph> of the queue.
		 */
		PACKAGE PacketBuffer removeFromQueue()
		{
			final PacketBuffer newHead;

			require(this.onAQueue());

			if (this.next == this)
			{
				// singleton list.
				newHead = null;
			}
			else
			{
				newHead = this.next;
				
				// not a singleton
				this.next.prev = this.prev;
				this.prev.next = this.next;

				// Leave the list in good shape
				ensure(newHead.onAQueue());
			}
			
			this.next = null;
			this.prev = null;

			ensure(this.notOnAQueue());
			return newHead;
		}
	

		PACKAGE void trade(PacketBuffer pb)
		{
			Object tmp;

			tmp = this.owner;
			this.owner = pb.owner;
			pb.owner = tmp;
		}
		
		PACKAGE synchronized void acquire()
		{
			this.users++;
		}

		PACKAGE synchronized void release()
		{
			if( this.users == 0 )
				throw new IllegalStateException("PacketBuffer released() one too many times.");
			this.users--;
			if( this.users == 0 )
				ProtocolLayer.recycle(this);
		}

		PACKAGE int getRefCount()
		{
			// XXX require(this.users > 0)?
			return this.users;
		}
		
		public String toString()
		{
			return "PacketBuffer[data="
				+ this.data
				+ "]";
		}
	}

	/**
	 * PacketHandler is an abstract class that is subclassed by the
	 * users of protocol layers to handle incoming packets.
	 */
	PACKAGE static abstract class PacketHandler
	{
		/**
		 * The source mask determines what incoming packets will be
		 * sent to the handle method.
		 */
		PACKAGE PacketAddressMask source;
		/**
		 * The method that will handle incoming packets.  The return
		 * value is taken as the replacement for the passed in
		 * packet.  If the handler doesn't have any extra packets
		 * to return to the protocol layer it should just return
		 * the passed in packet.
		 */
		PACKAGE abstract PacketBuffer handle(PacketBuffer pb);

		public String toString()
		{
			return "PacketHandler[source="
				+ this.source
				+ "]";
		}
	}

	/**
	 * PacketGenerator is an abstract class that is subclassed by the
	 * users of protocol layers to handle generating outgoing packets.
	 */
	PACKAGE static abstract class PacketGenerator
	{
		/**
		 * Function called by the protocol layer to get more packets
		 * to send.
		 */
		PACKAGE abstract PacketBuffer generate();
		/**
		 * The destination of any generated packets.
		 */
		PACKAGE PacketAddress destination;
		
		public String toString()
		{
			return "PacketGenerator[destination="
				+ this.destination
				+ "]";
		}
	}

	PACKAGE interface PacketRecycler
	{
		void recycle(PacketBuffer pb);
	}
}
