/*
 * UDPLayer.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;
import java.util.Hashtable;
import java.util.Vector;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.SocketException;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/**
 * UDPLayer is a protocol layer implementation that runs over UDP.
 */
PACKAGE class UDPLayer
	extends ProtocolLayer
{
	DatagramPacket dp = new DatagramPacket(new byte[0], 0);
	private DatagramSocket ds;
	
	PACKAGE UDPLayer()
		throws Exception
	{
		this.ds = new DatagramSocket();
	}

	public String toString()
	{
		return "UDPLayer[]";
	}

	/**
	 * UDP specific address implementation.
	 */
	PACKAGE static class UDPAddress
		implements ProtocolLayer.PacketAddress
	{
		PACKAGE InetAddress ia = null;
		PACKAGE int port = -1;

		/**
		 * Parses UDP address strings.  The format is:
		 *  "<src address>:<dest port>:<dest address>:<src port>
		 */
		public void parseAddress(String addr)
			throws PacketAddressFormatException
		{
			StringTokenizer st = new StringTokenizer(addr, ":");

			try
			{
				st.nextToken(); // srcAddr
				this.port = Integer.parseInt(st.nextToken());
				this.ia = InetAddress.
					getByName(st.nextToken());
			}
			catch(UnknownHostException e)
			{
				throw new PacketAddressFormatException();
			}
			catch(NoSuchElementException e)
			{
			}
			if( ia == null )
				throw new PacketAddressFormatException();
		}

		public String toString()
		{
			return "UDPAddress[ia="
				+ this.ia
				+ "; port="
				+ this.port
				+ "]";
		}
	}
	
	PACKAGE static class UDPAddressMask
		implements ProtocolLayer.PacketAddressMask
	{
		PACKAGE InetAddressMask srcMask;
		PACKAGE int srcPort = -1;
		PACKAGE int destPort = -1;
		PACKAGE InetAddressMask destMask;

		/**
		 * Parses UDP address mask strings.  The format is:
		 *  "<src address>:<dest port>:<dest address>:<src port>
		 */
		public void parseAddressMask(String addr)
			throws PacketAddressFormatException
		{
			StringTokenizer st = new StringTokenizer(addr, ":");
			String comp;
		
			try
			{
				this.srcMask =
					new InetAddressMask(st.nextToken());
				comp = st.nextToken();
				if( !comp.equals("*") )
					this.destPort = Integer.parseInt(comp);
				this.destMask =
					new InetAddressMask(st.nextToken());
				comp = st.nextToken();
				if( !comp.equals("*") )
					this.srcPort = Integer.parseInt(comp);
			}
			catch(UnknownHostException e)
			{
				throw new PacketAddressFormatException();
			}
			catch(NoSuchElementException e)
			{
			}
		}

		/**
		 * Check for a match between our mask and the datagram packet
		 */
		PACKAGE boolean matchPacket(DatagramPacket dp)
		{
			boolean retval = false;

			if( this.srcMask.match(dp.getAddress()) &&
			    ((this.srcPort == -1) ||
			     (dp.getPort() == this.srcPort)) )
			{
				retval = true;
			}
			return retval;
		}

		PACKAGE int priority()
		{
			int retval;

			retval = this.destMask.priority() +
				this.srcMask.priority();
			if( this.srcPort != -1 )
				retval++;
			if( this.destPort != -1 )
				retval++;
			return retval;
		}
		
		public int hashCode()
		{
			return this.srcMask.hashCode() +
				this.srcPort +
				this.destPort +
				this.destMask.hashCode();
		}
		
		public boolean equals(Object o)
		{
			boolean retval = false;
			
			if( o instanceof UDPAddressMask )
			{
				UDPAddressMask uam = (UDPAddressMask)o;

				if( uam.srcMask.equals(this.srcMask) &&
				    uam.destMask.equals(this.destMask) &&
				    (uam.srcPort == this.srcPort) &&
				    (uam.destPort == this.destPort) )
				{
					retval = true;
				}
			}
			return( retval );
		}
		
		public String toString()
		{
			return "UDPAddressMask[srcMask="
				+ this.srcMask
				+ "; srcPort="
				+ this.srcPort
				+ "; destPort="
				+ this.destPort
				+ "; destMask="
				+ this.destMask
				+ "]";
		}
	}

	/**
	 * The table of sockets that we have open for reading, organized by
	 * port number.
	 */
	private final Hashtable sockets = new Hashtable();

	/**
	 * Class that takes care of reading sockets and calling any associated
	 * packet handlers.
	 */
	private static class SocketReader
		implements Runnable
	{
		/**
		 * Vector of packet handlers attached to this socket.
		 */
		private Vector masks = new Vector();
		private DatagramSocket ds;
		private int users = 1;

		PACKAGE SocketReader(int port)
			throws SocketException
		{
			this.ds = new DatagramSocket(port);
		}

		/**
		 * Find a packet handler that matches the given packet, we
		 * just check the source masks of each handler.
		 */
		private PacketHandler matchMask(DatagramPacket dp)
		{
			PacketHandler ph, retval = null;
			int size = this.masks.size();
			int lpc;
			
			for( lpc = 0; (lpc < size) && (retval == null); lpc++ )
			{
				ph = (PacketHandler)this.masks.elementAt(lpc);
				if( ((UDPAddressMask)ph.source).
				    matchPacket(dp) )
				{
					retval = ph;
				}
			}
			return retval;
		}

		/**
		 * Add a packet handler to this socket.
		 */
		PACKAGE void addHandler(PacketHandler ph)
		{
			int lpc, size;
			
			synchronized(this.masks)
			{
				size = this.masks.size();
				for( lpc = 0; lpc < size; lpc++ )
				{
					PacketHandler elem =
						(PacketHandler)this.masks.
						elementAt(lpc);

					if( ((UDPAddressMask)ph.source).
					    priority() >=
					    ((UDPAddressMask)elem.source).
					    priority() )
						break;
				}
				this.masks.insertElementAt(ph, lpc);
				if( size != 0 )
					this.users++;
			}
		}

		/**
		 * Remove a packet handler from this socket.
		 */
		PACKAGE boolean remHandler(PacketHandler ph)
		{
			boolean retval = false;
			
			synchronized(this.masks)
			{
				this.users--;
				this.masks.removeElement(ph);
				if( this.masks.size() == 0 )
				{
					DatagramPacket dp;

					retval = true;
					try
					{
						dp = new DatagramPacket(
							new byte[0],
							0,
							InetAddress.
							getLocalHost(),
							this.ds.
							getLocalPort());
						this.ds.send(dp);
					}
					catch(Exception e)
					{
						e.printStackTrace();
					}
				}
			}
			return( retval );
		}
		
		public void run()
		{
			PacketBuffer pb = new PacketBuffer(null);
			DatagramPacket dp;
			PacketHandler ph;

			dp = new DatagramPacket(pb.data, pb.data.length);
			while( this.users != 0 )
			{
				try
				{
					this.ds.receive(dp);
					pb.length = dp.getLength();
					TRACE(Trace.NODEOS_INCOMING, "in!");
					TRACE(Trace.NODEOS_INCOMING, "received: " + Debug.dbgPacketSummary(pb));
					synchronized(this.masks)
					{
						ph = this.matchMask(dp);
					}
					if( ph != null )
					{
						pb = ph.handle(pb);
						dp.setData(pb.data);
						dp.setLength(pb.data.length);
					}
				}
				catch(Exception e)
				{
					Node.println("exception!! " + e);
					e.printStackTrace();
				}
			}
			TRACE(Trace.NODEOS_INCOMING, "recv port (on " +this.ds.getLocalPort()+ "): exiting");
			this.ds.close();
		}
	}

	PACKAGE String getID()
	{
		return "udp";
	}

	PACKAGE int getPacketLength()
	{
		return 16384;
	}

	PACKAGE PacketAddress createAddress()
	{
		return new UDPAddress();
	}

	PACKAGE PacketAddressMask createAddressMask()
	{
		return new UDPAddressMask();
	}

	PACKAGE void attachHandler(PacketHandler ph)
		throws IOException
	{
		int port = ((UDPAddressMask)ph.source).destPort;
		Integer i = new Integer(port);
		SocketReader sr;

		synchronized(this.sockets)
		{
			if( (sr = (SocketReader)this.sockets.
			     get(i)) == null )
			{
				java.lang.Thread srThread;
				
				sr = new SocketReader(port);
				srThread = new java.lang.Thread(sr);
				srThread.setName("SocketReader for UDPLayer "
						 + "port "
						 + port);
				srThread.start();
				this.sockets.put(i, sr);
			}
			sr.addHandler(ph);
		}
	}

	PACKAGE void detachHandler(PacketHandler ph)
	{
		int port = ((UDPAddressMask)ph.source).destPort;
		Integer i = new Integer(port);
		SocketReader sr;

		try
		{
			synchronized(this.sockets)
			{
				if( (sr = (SocketReader)this.sockets.get(i)) != null )
				{
					if( sr.remHandler(ph) )
						this.sockets.remove(i);
				}
			}
		}
		catch(Exception e)
		{
		}
	}

	PACKAGE void attachGenerator(PacketGenerator pg)
	{
		/* We don't need to track the generators */
	}

	PACKAGE void generatorReady(PacketGenerator pg)
	{
		PacketBuffer pb;
		UDPAddress ua;

		/* Get the destination and setup the datagram packet */
		ua = (UDPAddress)pg.destination;
		synchronized(this.dp)
		{
			this.dp.setAddress(ua.ia);
			this.dp.setPort(ua.port);
			pb = pg.generate();
			this.dp.setData(pb.data);
			this.dp.setLength(pb.length);
			try
			{
				this.ds.send(this.dp);
			}
			catch(IOException e)
			{
			}
		}
	}

	PACKAGE void detachGenerator(PacketGenerator pg)
	{
		/* We don't need to track the generators */
	}
}
